import logging
import traceback

from mycroft.common import aws_s3


class org_num_normalizer:
    def __init__(
        self,
        recommendation_types,
        prevalence_thresh,
        s3_bucket,
        hash_pool_key,
        cert_pool_key,
        it_tool_pool_key,
    ):
        self.thresholds = dict()
        self.aws_data = aws_s3.AwsS3({})
        for recommendation_type in recommendation_types:
            low_threshold, high_threshold = self.gen_thresholds(
                recommendation_type,
                prevalence_thresh,
                s3_bucket,
                hash_pool_key,
                cert_pool_key,
                it_tool_pool_key,
            )
            self.thresholds[recommendation_type] = (low_threshold, high_threshold)

    def gen_it_tool_rule_pool(self, prevalence_thresh, s3_bucket, pool_key):
        pool_df = self.aws_data.get_dataframe_from_s3(
            s3_bucket, pool_key, col_names=["rule_path", "org_count", "value2"]
        )
        # prevalence_thresh filter
        pool_df = pool_df[
            (pool_df.filter(items=["org_count"]) > prevalence_thresh).any(axis=1)
        ]
        pool_df = pool_df.filter(items=["rule_path", "org_count"])
        return pool_df.sort_values(by=["org_count"], ascending=False)

    # def gen_permission_rule_pool(
    #     self, prevalence_thresh, database, policy_table, force_query
    # ):
    #     query = f"""
    #     WITH t1 AS
    #         (SELECT lower(application_value) application_value,
    #                 count(DISTINCT org_id) org_cnt
    #         FROM {database}.{policy_table}
    #         WHERE application_value LIKE '%\\%' AND
    #             (action = 'ALLOW' OR action = 'IGNORE')
    #         GROUP BY 1
    #         ORDER BY 2 DESC)
    #     SELECT t1.* FROM t1 WHERE t1.org_cnt > {prevalence_thresh}"""
    #     return self.aws_data.get_dataframe(
    #         gen_signature(query, "policy-automation"), "temp.csv", force_query
    #     )

    def gen_hash_rule_pool(self, prevalence_thresh, s3_bucket, pool_key):
        pool_df = self.aws_data.get_dataframe_from_s3(s3_bucket, pool_key)
        # prevalence_thresh filter
        pool_df = pool_df[
            (pool_df.filter(items=["org_count"]) > prevalence_thresh).any(axis=1)
        ]
        pool_df = pool_df.filter(items=["proc_sha256hash", "org_count"])
        return pool_df.sort_values(by=["org_count"], ascending=False)

    def gen_cert_rule_pool(self, prevalence_thresh, s3_bucket, pool_key):
        pool_df = self.aws_data.get_dataframe_from_s3(s3_bucket, pool_key)
        # trim whitespace
        pool_df["cert_signedby"] = pool_df["cert_signedby"].str.strip()
        pool_df["cert_authority"] = pool_df["cert_authority"].str.strip()
        # concat
        pool_df["cert_name"] = pool_df[["cert_signedby", "cert_authority"]].agg(
            " + ".join, axis=1
        )
        # prevalence_thresh filter
        pool_df = pool_df[
            (pool_df.filter(items=["org_count"]) > prevalence_thresh).any(axis=1)
        ]
        pool_df = pool_df.filter(items=["cert_name", "org_count"])
        return pool_df.sort_values(by=["org_count"], ascending=False)

    def normalize_pool(self, df):
        rows = df.shape[0]
        if rows > 1:
            high_threshold = df.iloc[rows // 3]["org_count"]
            low_threshold = df.iloc[rows * 2 // 3]["org_count"]
        else:
            high_threshold = 0
            low_threshold = 0
        return low_threshold, high_threshold

    def gen_thresholds(
        self,
        recommendation_type,
        prevalence_thresh,
        s3_bucket,
        hash_pool_key,
        cert_pool_key,
        it_tool_pool_key,
    ):
        try:
            if recommendation_type == "it_tools":
                df = self.gen_it_tool_rule_pool(
                    prevalence_thresh, s3_bucket, it_tool_pool_key
                )
            # elif recommendation_type == "perm_rule":
            #     df = self.gen_permission_rule_pool(
            #         prevalence_thresh, s3_bucket,
            #     )
            elif recommendation_type == "hash":
                df = self.gen_hash_rule_pool(
                    prevalence_thresh, s3_bucket, hash_pool_key
                )
            elif recommendation_type == "cert":
                df = self.gen_cert_rule_pool(
                    prevalence_thresh, s3_bucket, cert_pool_key
                )
            else:
                raise Exception(
                    f"recommendation_type [{recommendation_type}] error. Type values: it_tools, perm_rule, "
                    f"proc_sha256hash, cert"
                )
            return self.normalize_pool(df)
        except Exception as e:
            traceback.print_exc()
            logging.warning("Error in generate threshold: %s", str(e))
            return None, None

    def get_normalized_org_cnt(self, org_cnt, recommendation_type):
        low_threshold, high_threshold = self.thresholds[recommendation_type]
        if low_threshold and high_threshold and high_threshold > low_threshold:
            if org_cnt > high_threshold:
                return 2
            elif low_threshold < org_cnt <= high_threshold:
                return 1
            else:
                return 0
        elif low_threshold == high_threshold:
            return 0
        else:
            logging.error("Error in generated thresholds")
            return None
