import json
import os

import boto3
import numpy as np
import pandas as pd

from mycroft.common import aws_s3, gen_signature
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def get_org_device_count_dataframe(start_time, end_time, aws_data, force_query):
    """
    get_org_device_count_dataframe:
    A Function to query org - device count in event table between start_time and end_time
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""
    SELECT env, org_id AS orgid, CARDINALITY(merge(CAST(alert_device_count_hll_sketch AS HYPERLOGLOG))) AS device_count
    FROM proc_hash_cert_alert_db.proc_hash_cert_daily_alert_dev_count where event_date between '{start_time}' and '{end_time}'
    GROUP BY 1,2
    """
    logger.debug(query)
    org_device_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_device_count.csv",
        force_query,
    )
    return org_device_count_df


def get_org_whitelist_count_dataframe(
    aws_data, policy_override_table, database, force_query
):
    """
    Function to query org - whitelist rule count in reputation table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""select env, orgid, count(*) AS reputation_whitelist_rule_count
        from {database}.{policy_override_table}
        where authlisttype = 'WHITE_LIST'
        group by env, orgid"""
    logger.debug(query)
    org_whitelist_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_whitelist_count.csv",
        force_query,
    )
    return org_whitelist_count_df


def get_org_blacklist_count_dataframe(
    aws_data, policy_override_table, database, force_query
):
    """
    Function to query org - blacklist rule count in reputation table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""select env, orgid, count(*) AS reputation_blacklist_rule_count
        from {database}.{policy_override_table}
        where authlisttype = 'BLACK_LIST'
        group by env, orgid"""
    logger.debug(query)
    org_blacklist_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_blacklist_count.csv",
        force_query,
    )
    return org_blacklist_count_df


def get_org_existence_time_dataframe(aws_data, force_query):
    """
    Function to query org - org existence time in alert table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = """SELECT env, org_id AS orgid, date_diff('day', min(date(event_date)), max(date(event_date))) org_existence_in_days
    FROM proc_hash_cert_alert_db.proc_hash_cert_daily_alert_dev_count group by 1, 2"""
    logger.debug(query)
    org_existence_time_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_existence_time.csv",
        force_query,
    )
    return org_existence_time_df


def get_org_policy_permission_rule_count_dataframe(
    database, policy_table, aws_data, force_query
):
    """
    Function to query org - permission rule count in policy table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""select env, org_id AS orgid, device_group_id, count(*) AS permission_rule_count
        from {database}.{policy_table} where action in ('ALLOW','IGNORE') group by env, org_id, device_group_id"""
    logger.debug(query)
    org_policy_permission_rule_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_policy_permission_rule_count.csv",
        force_query,
    )
    return org_policy_permission_rule_count_df


"""
Function to query org - blocking rule count in policy table
Store the result in S3
Pandas dataframe read the result from S3
"""


def get_org_policy_blocking_rule_count_dataframe(
    database, policy_table, aws_data, force_query
):
    query = f"""select env, org_id AS orgid, device_group_id, count(*) AS blocking_rule_count
        from {database}.{policy_table} where action in ('TERMINATE','DENY') group by env, org_id, device_group_id"""
    logger.debug(query)
    org_policy_blocking_rule_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_policy_blocking_rule_count.csv",
        force_query,
    )
    return org_policy_blocking_rule_count_df


def get_org_blocking_event_count_dataframe(start_time, end_time, aws_data, force_query):
    """
    Function to query org - event count between start_time and end_time in event table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""
        SELECT env, org_id AS orgid, sum(alert_cnt) AS blocking_event_count
        FROM  proc_hash_cert_alert_db.proc_hash_cert_daily_alert_dev_count
        WHERE event_date between '{start_time}' AND '{end_time}'
        GROUP BY 1, 2
        """
    logger.debug(query)
    org_blocking_event_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_blocking_event_count.csv",
        force_query,
    )
    return org_blocking_event_count_df


def get_org_user_added_rule_count_df(
    database, default_rule_str, policy_table, aws_data, force_query
):
    """
    Function to query org - rule list in policy table
    Store the result in S3
    Pandas dataframe read the result from S3
    """
    query = f"""WITH t AS (select env, org_id AS orgid, application_value, operation, action
             from {database}.{policy_table} WHERE concat(application_value, '^', operation, '^', action) NOT IN ({default_rule_str}))
             SELECT env, orgid, action, count(*) AS cnt FROM t GROUP BY env, orgid, action
            """
    logger.debug(query)
    org_policy_permission_rule_count_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "org_rules.csv",
        force_query,
    )
    return org_policy_permission_rule_count_df


"""
Function parse json policy file and extract rules in it
"""


def read_json_to_rules(json_data):
    default_rule_set = set()
    for rule in json_data["rules"]:
        val = (
            rule["application"]["value"]
            + "^"
            + rule["action"]
            + "^"
            + rule["operation"]
        )
        logger.debug(f"default rule: {val}")
        default_rule_set.add(val)
    return default_rule_set


"""
Function to read the folder which stores the policy files
Extract rules in each policy json file
Store the rules in the default_rule_set
"""


def read_json_to_default_rule_set(aws_s3_policy_json_bucket, aws_s3_policy_json_prefix):
    default_rule_set = set()
    s3 = boto3.resource("s3")
    bucket = s3.Bucket(aws_s3_policy_json_bucket)

    for obj in bucket.objects.filter(Prefix=aws_s3_policy_json_prefix):
        if obj.key.endswith(".json"):
            rule_subset = read_json_to_rules(json.load(obj.get()["Body"]))
            default_rule_set = default_rule_set.union(rule_subset)
    return default_rule_set


"""
Function to get org - user add rule count
user added rule is defined as the following:
The rule is not belonging to any default policy
"""


def get_user_added_rule_count_dataframe(
    aws_data,
    policy_table,
    force_query,
    aws_s3_policy_json_bucket,
    aws_s3_policy_json_prefix,
    params,
):
    default_rule_set = read_json_to_default_rule_set(
        aws_s3_policy_json_bucket, aws_s3_policy_json_prefix
    )
    default_rule_str = ",".join(
        ["'" + default_rule + "'" for default_rule in default_rule_set]
    )
    logger.debug(f"default rule set size: {len(default_rule_set)}")
    org_user_added_rule_count_df = get_org_user_added_rule_count_df(
        params["database"], default_rule_str, policy_table, aws_data, force_query
    )
    org_user_added_permission_rule_count_df = (
        org_user_added_rule_count_df[
            org_user_added_rule_count_df["action"].isin(["ALLOW", "IGNORE"])
        ]
        .drop(["action"], axis=1)
        .groupby(["env", "orgid"])
        .sum()
        .reset_index()
    )
    org_user_added_permission_rule_count_df.rename(
        columns={"cnt": "user_added_permission_rule_count"}, inplace=True
    )
    org_user_added_blocking_rule_count_df = (
        org_user_added_rule_count_df[
            org_user_added_rule_count_df["action"].isin(["DENY", "TERMINATE"])
        ]
        .drop(["action"], axis=1)
        .groupby(["env", "orgid"])
        .sum()
        .reset_index()
    )
    org_user_added_blocking_rule_count_df.rename(
        columns={"cnt": "user_added_blocking_rule_count"}, inplace=True
    )
    org_user_added_total_rule_count_df = (
        org_user_added_rule_count_df.drop(["action"], axis=1)
        .groupby(["env", "orgid"])
        .sum()
        .reset_index()
    )
    org_user_added_total_rule_count_df.rename(
        columns={"cnt": "user_added_rule_count"}, inplace=True
    )
    org_user_added_rule_summary_df = org_user_added_total_rule_count_df.merge(
        org_user_added_blocking_rule_count_df, on=["env", "orgid"], how="outer"
    )
    org_user_added_rule_summary_df = org_user_added_rule_summary_df.merge(
        org_user_added_permission_rule_count_df, on=["env", "orgid"], how="outer"
    )
    org_user_added_rule_summary_df.fillna(0, inplace=True)
    return org_user_added_rule_summary_df


"""
Get org - total event count between start_time and end_time
"""


def get_org_total_event_count_dataframe(
    envs, start_time, end_time, aws_data, force_query
):
    org_total_event_count_dfs = []
    for env in envs:
        query = f"""
            SELECT environmentname AS env, CAST(organizationid AS Integer) AS orgid, count(*) AS all_event_count
            FROM athena.flink_{env}
            WHERE event_date between '{start_time}' and '{end_time}' AND try(CAST(organizationid AS INTEGER)) IS NOT NULL
            GROUP BY 1, 2
            """
        logger.debug(query)
        org_total_event_count_df = aws_data.get_dataframe(
            gen_signature(query, "policy-automation"),
            "org_total_event_count.csv",
            force_query,
        )
        org_total_event_count_dfs.append(org_total_event_count_df)
    return pd.concat(org_total_event_count_dfs)


"""
Get all org id list
"""


def get_all_org_id_dataframe(database, policy_table, aws_data, force_query):
    query = f"""
    SELECT distinct env, org_id AS orgid
    FROM {database}.{policy_table}
    """
    logger.debug(query)
    org_maturity_df = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "all_orgs.csv",
        force_query,
    )
    return org_maturity_df


def get_scaled_column(df, feature):
    min_val = min(df[feature])
    max_val = max(df[feature])
    return 100 * (df[feature] - min_val) / (max_val - min_val)


"""
Upload a file to S3
"""


def upload_file(file_name, bucket, object_name=None):
    """Upload a file to an S3 bucket

    :param file_name: File to upload
    :param bucket: Bucket to upload to
    :param object_name: S3 object name. If not specified then file_name is used
    :return: True if file was uploaded, else False
    """

    # If S3 object_name was not specified, use file_name
    if object_name is None:
        object_name = file_name

    # Upload the file
    s3_client = boto3.client("s3")
    try:
        s3_client.upload_file(file_name, bucket, object_name)
    except Exception as e:
        print(e)
        return False
    return True


"""
Main function to call and generate org_maturity_score
params = {
    'region': 'us-east-1',
    'database': 'tau_cluster_scratch',
    'bucket': 'cdca-us-east-1-cerberus-scratch',
    'outputBucket': 'cdca-us-east-1-cerberus-scratch',
    'outputFolder': 'zmo',
    'outputLocation': 's3://cdca-us-east-1-cerberus-scratch/zmo/'
}
"""


def generate_org_maturity_list_dataframe(
    params,
    envs,
    start_time,
    end_time,
    policy_override_table,
    policy_table,
    # aws_s3_policy_json_bucket="cdca-prod-us-east-1-intermediate-data",
    aws_s3_policy_json_bucket="gdev4-policysuggestion-infra-intermediate-data-bucket",
    aws_s3_policy_json_prefix="policy_json_files",
    force_query=True,
):
    """
    THIS METHOD IS NOT USED BY EXISTING PIPELINE
    """
    # ORG_MATURITY_SCORE1 is calculated on "device_count", "user_added_rule_count" and "blocking_event_ratio"
    ORG_MATURITY_SCORE1_THRESHOLD = os.getenv("ORG_MATURITY_SCORE1_THRESHOLD", 30)
    # ORG_MATURITY_SCORE2 is calculated on "device_count", "ORG_MATURITY_SCORE2" and "reputation_whitelist_rule_count"
    ORG_MATURITY_SCORE2_THRESHOLD = os.getenv("ORG_MATURITY_SCORE2_THRESHOLD", None)

    aws_data = aws_s3.AwsS3(params)

    # get org device count dataframe#
    org_device_count_df = get_org_device_count_dataframe(
        start_time, end_time, aws_data, force_query
    )
    logger.debug(f"get org device count dataframe. size: {org_device_count_df.shape}")
    # get org reputation whitelist count dataframe#
    org_whitelist_count_df = get_org_whitelist_count_dataframe(
        aws_data, policy_override_table, params["database"], force_query
    )
    logger.debug(
        f"get org reputation whitelist count dataframe. size: {org_whitelist_count_df.shape}"
    )
    # get org reputation blacklist count dataframe#
    org_blacklist_count_df = get_org_blacklist_count_dataframe(
        aws_data, policy_override_table, params["database"], force_query
    )
    logger.debug(
        f"get org reputation blacklist count dataframe. size: {org_blacklist_count_df.shape}"
    )
    # get org policy permission rule count dataframe#
    org_policy_permission_rule_count_df = (
        get_org_policy_permission_rule_count_dataframe(
            params["database"], policy_table, aws_data, force_query
        )
    )
    logger.debug(
        f"get org policy permission rule count dataframe. size: {org_policy_permission_rule_count_df.shape}"
    )
    # get the median of permission rule count for policies dataframe#
    policy_permission_rule_count_median = org_policy_permission_rule_count_df[
        "permission_rule_count"
    ].median()
    logger.debug(
        f"median permission rule per policy: {policy_permission_rule_count_median}"
    )

    org_policy_blocking_rule_count_df = get_org_policy_blocking_rule_count_dataframe(
        params["database"], policy_table, aws_data, force_query
    )
    policy_blocking_rule_count_median = org_policy_blocking_rule_count_df[
        "blocking_rule_count"
    ].median()
    logger.debug(f"median permission rule: {policy_blocking_rule_count_median}")

    org_policy_permission_blocking_rule_count_df = (
        org_policy_blocking_rule_count_df.merge(
            org_policy_permission_rule_count_df,
            on=["env", "orgid", "device_group_id"],
            how="outer",
        )
    )
    logger.debug(
        f"get org policy permission blocking rule count df: {org_policy_permission_blocking_rule_count_df.shape}"
    )
    org_policy_permission_blocking_rule_count_df.fillna(0, inplace=True)

    org_policy_permission_blocking_rule_count_df[
        "exist_X_blocking_Y_permission_rule"
    ] = (
        org_policy_permission_blocking_rule_count_df["blocking_rule_count"]
        >= policy_blocking_rule_count_median
    ) & (
        org_policy_permission_blocking_rule_count_df["permission_rule_count"]
        >= policy_permission_rule_count_median
    )

    org_policy_permission_blocking_rule_count_df[
        "num_of_policy_with_more_than_median_blocking_permission_rule"
    ] = org_policy_permission_blocking_rule_count_df[
        "exist_X_blocking_Y_permission_rule"
    ].astype(
        int
    )

    org_exist_X_permission_Y_blocking_df = (
        org_policy_permission_blocking_rule_count_df.drop(
            ["exist_X_blocking_Y_permission_rule"], axis=1
        )
    )
    org_num_X_permission_Y_blocking_df = (
        org_exist_X_permission_Y_blocking_df.groupby(["env", "orgid"])
        .agg(
            {
                "num_of_policy_with_more_than_median_blocking_permission_rule": "sum",
                "device_group_id": "size",
                "blocking_rule_count": "sum",
                "permission_rule_count": "sum",
            }
        )
        .reset_index()
    )
    org_num_X_permission_Y_blocking_df.rename(
        columns={"device_group_id": "total_policy_num"}, inplace=True
    )

    org_user_added_rules_count_df = get_user_added_rule_count_dataframe(
        aws_data,
        policy_table,
        force_query,
        aws_s3_policy_json_bucket,
        aws_s3_policy_json_prefix,
        params,
    )
    logger.debug(
        f"get org_user_added_rules_count_df. size: {org_user_added_rules_count_df.shape}"
    )
    org_exitence_time_df = get_org_existence_time_dataframe(aws_data, force_query)
    logger.debug(f"get org_exitence_time_df. size: {org_exitence_time_df.shape}")
    org_blocking_event_count_df = get_org_blocking_event_count_dataframe(
        start_time, end_time, aws_data, force_query
    )
    logger.debug(
        f"get org_blocking_event_count_df. size: {org_blocking_event_count_df.shape}"
    )
    org_total_event_count_df = get_org_total_event_count_dataframe(
        envs, start_time, end_time, aws_data, force_query
    )
    logger.debug(
        f"get org_total_event_count_df. size: {org_total_event_count_df.shape}"
    )
    org_event_ratio_df = org_total_event_count_df.merge(
        org_blocking_event_count_df, on=["env", "orgid"], how="outer"
    )
    org_event_ratio_df["blocking_event_ratio"] = (
        org_event_ratio_df["blocking_event_count"]
        / org_event_ratio_df["all_event_count"]
    )
    org_event_ratio_df.fillna(0, inplace=True)

    org_maturity_df = get_all_org_id_dataframe(
        params["database"], policy_table, aws_data, force_query
    )

    org_maturity_df = org_maturity_df.merge(
        org_device_count_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df = org_maturity_df.merge(
        org_whitelist_count_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df = org_maturity_df.merge(
        org_blacklist_count_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df = org_maturity_df.merge(
        org_exitence_time_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df = org_maturity_df.merge(
        org_user_added_rules_count_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df = org_maturity_df.merge(
        org_num_X_permission_Y_blocking_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df.fillna(0, inplace=True)
    org_maturity_df = org_maturity_df.merge(
        org_event_ratio_df, on=["env", "orgid"], how="left"
    )
    org_maturity_df.fillna(0, inplace=True)
    cols = org_maturity_df.columns.to_list()
    cols.remove("blocking_event_ratio")
    cols.remove("env")
    logger.debug(f"cols: {cols}")
    org_maturity_df[cols] = org_maturity_df[cols].applymap(np.int64)

    org_maturity_df["blocking_event_ratio"].mask(
        org_maturity_df["all_event_count"] == 0, None, inplace=True
    )
    org_maturity_score_df = org_maturity_df.drop(
        org_maturity_df[org_maturity_df["all_event_count"] == 0].index
    )
    org_maturity_score_df["org_maturity_score1"] = (
        get_scaled_column(org_maturity_score_df, "device_count")
        + get_scaled_column(org_maturity_score_df, "user_added_rule_count")
        + 100
        - get_scaled_column(org_maturity_score_df, "blocking_event_ratio")
    ) / 3
    org_maturity_score_df["org_maturity_score2"] = (
        get_scaled_column(org_maturity_score_df, "device_count")
        + get_scaled_column(org_maturity_score_df, "user_added_rule_count")
        + get_scaled_column(org_maturity_score_df, "reputation_whitelist_rule_count")
    ) / 3
    if ORG_MATURITY_SCORE1_THRESHOLD:
        org_maturity_score_df = org_maturity_score_df[
            org_maturity_score_df["org_maturity_score1"]
            >= ORG_MATURITY_SCORE1_THRESHOLD
        ]
    if ORG_MATURITY_SCORE2_THRESHOLD:
        org_maturity_score_df = org_maturity_score_df[
            org_maturity_score_df["org_maturity_score2"]
            >= ORG_MATURITY_SCORE2_THRESHOLD
        ]
    return org_maturity_df, org_maturity_score_df
