import os

from mycroft.common import aws_s3, gen_signature
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def get_permission_rules(aws_data, force_query):
    query = "select device_group_id, lower(application_value) AS application_value, operation, action \
         from tau_cluster_scratch.policy \
         where application_value like '%\\%' \
         and (length(application_value) - length(replace(application_value, '**', ''))) < 10 \
         and not application_value like '%;%' \
         and action in ('ALLOW','IGNORE')"
    return aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "all_permission_rules.csv",
        force_query,
    )


def get_policy_device_count(aws_data, force_query):
    query = "select device_group_id, count(distinct device_id) AS device_count \
         from tau_cluster_scratch.policy_device_mapping \
         group by device_group_id"
    return aws_data.get_dataframe(
        gen_signature(query, "policy-automation"),
        "policy_device_count.csv",
        force_query,
    )


def get_scaled_column(df, feature):
    min_val = min(df[feature])
    max_val = max(df[feature])
    return 100 * (df[feature] - min_val) / (max_val - min_val)


def generate_permission_rule_with_prevalence_df(aws_data, force_query):
    device_count_ratio = 0.75
    rule_count_ratio = 1.0 - device_count_ratio
    logger.info("get all permission rules dataframe")
    all_data_df = get_permission_rules(aws_data, force_query)
    logger.info("get policy device count dataframe")
    policy_device_count_df = get_policy_device_count(aws_data, force_query)
    logger.info("get policy rule count dataframe")
    policy_rule_count_df = (
        all_data_df.drop(["operation", "action"], axis=1)
        .groupby(["device_group_id"])
        .size()
        .to_frame("rule_count")
        .reset_index()
    )
    logger.info("Join and get policy rule/device count dataframe")
    policy_rule_device_df = policy_rule_count_df.merge(
        policy_device_count_df, on="device_group_id", how="outer"
    )
    policy_rule_device_df.fillna(0, inplace=True)
    logger.info("calculate prevalence")
    policy_rule_mean = policy_rule_device_df["rule_count"].mean()
    policy_device_mean = policy_rule_device_df["device_count"].mean()
    policy_rule_device_df["prevalent_rate"] = (
        device_count_ratio * policy_rule_device_df["device_count"] / policy_device_mean
        + rule_count_ratio * policy_rule_device_df["rule_count"] / policy_rule_mean
    )
    total_weight = policy_rule_device_df["prevalent_rate"].sum()
    policy_rule_device_df["prevalent_rate"] = (
        policy_rule_device_df["prevalent_rate"] / total_weight
    )
    policy_rule_device_df["prevalent_rate"] = get_scaled_column(
        policy_rule_device_df, "prevalent_rate"
    )
    df = all_data_df.merge(policy_rule_device_df, on="device_group_id", how="outer")
    df = df.drop(["device_group_id", "rule_count"], axis=1)
    df = df.groupby(["application_value", "operation", "action"]).sum().reset_index()
    df = df.sort_values(by=["device_count"], ascending=False)
    return df


"""
Main function to call
params = {
    'region': 'us-east-1',
    'database': 'tau_cluster_scratch',
    'bucket': 'cdca-us-east-1-cerberus-scratch',
    'outputBucket': 'cdca-us-east-1-cerberus-scratch',
    'outputFolder': 'zmo',
    'outputLocation': 's3://cdca-us-east-1-cerberus-scratch/zmo/'
}
"""


def generate_permission_rule_list(params, output_filename, force_query=True):
    """
    THIS METHOD IS NOT USED BY EXISTING PIPELINE
    """
    aws_data = aws_s3.AwsS3(params)
    permission_rule_with_prevalence_df = generate_permission_rule_with_prevalence_df(
        aws_data, force_query
    )
    logger.info("write result to: %s", output_filename)
    permission_rule_with_prevalence_df.to_csv(output_filename, index=False)
    aws_data.upload_file(
        output_filename,
        "cdca-us-east-1-cerberus-scratch",
        "zmo/{}".format(output_filename),
    )
