import re
from datetime import datetime
from getpass import getuser
from json import dumps

from wcmatch import glob, pathlib


def validate_date(date_string: str):
    try:
        datetime.strptime(date_string, "%Y-%m-%d")
    except ValueError:
        raise ValueError(f"couldn't parse date ({date_string}), should be YYYY-MM-DD")


def translate(pattern: str) -> str:
    """[summary]
    A transation function to remove unsupported expression:
    Remove triple wildcards (***) - turn it into "**" instead.
    Make sure double wildcard (**) is preceded by posix style directory separator "/"

    Args:
        pattern (str): [a regular expression for a permission rule that needs tranlation]

    Returns:
        str: a valid regular expression for file path based rule.
    """
    pattern = pattern.replace("***", "**")
    pattern = re.sub(r"([a-zA-Z]+)(\*\*)", r"\1/\2", pattern)
    return pattern


def match(pattern: str, path: str):
    windows = False
    if "\\" in path:
        windows = True
        # We are parsing the path so we can convert it to posix path which is how python matching works
        parsed_path = pathlib.PureWindowsPath(path)
        # we are manully doing the same conversion for the search pattern.
        pattern = pattern.replace("\\", "/")
    else:
        parsed_path = pathlib.PurePosixPath(path)
    pattern = translate(pattern)
    if not windows:
        return glob.globmatch(parsed_path.as_posix(), pattern, flags=glob.GLOBSTAR)
    else:
        return glob.globmatch(
            parsed_path.as_posix(), pattern, flags=glob.GLOBSTAR | glob.IGNORECASE
        )


def gen_signature(query: str, project: str) -> str:
    now = datetime.now().replace(microsecond=0).isoformat()
    whoami = getuser()
    sig = {"user": whoami, "time": now, "project": project, "module_name": __name__}
    query = (
        (
            f"""
/*
{dumps(sig)}
*/
        """
        )
        + query
    )
    return query
