import io
import os
import re
import time
import traceback
from io import StringIO
from urllib.parse import urlparse

import boto3
import botocore
import pandas as pd
from pyathena import connect

from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def get_s3_client(cross_account_arn=None):
    if cross_account_arn:
        # Create the STS client
        sts_client = boto3.client("sts")

        # Assume the role
        assumed_role_object = sts_client.assume_role(
            RoleArn=cross_account_arn, RoleSessionName="policySuggestionMSC1AssumeRole"
        )

        # Extract the temporary credentials
        credentials = assumed_role_object["Credentials"]

        # Use the temporary credentials to create an S3 client
        return boto3.client(
            "s3",
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
    else:
        return boto3.client("s3")


def get_s3_resource(cross_account_arn=None):
    if cross_account_arn:
        # Create the STS client
        sts_client = boto3.client("sts")

        # Assume the role
        assumed_role_object = sts_client.assume_role(
            RoleArn=cross_account_arn, RoleSessionName="policySuggestionMSC1AssumeRole"
        )

        # Extract the temporary credentials
        credentials = assumed_role_object["Credentials"]

        # Use the temporary credentials to create an S3 resource
        return boto3.resource(
            "s3",
            aws_access_key_id=credentials["AccessKeyId"],
            aws_secret_access_key=credentials["SecretAccessKey"],
            aws_session_token=credentials["SessionToken"],
        )
    else:
        return boto3.resource("s3")


def assume_role_session(role_arn, session_name):
    # Assume a role using STS
    sts_client = boto3.client("sts")
    assumed_role_object = sts_client.assume_role(
        RoleArn=role_arn, RoleSessionName=session_name
    )

    credentials = assumed_role_object["Credentials"]

    # Construct session with assumed role's credentials
    session = boto3.Session(
        aws_access_key_id=credentials["AccessKeyId"],
        aws_secret_access_key=credentials["SecretAccessKey"],
        aws_session_token=credentials["SessionToken"],
    )
    return session


def upload_dataframe_to_s3(df, key: str, s3_client):
    csv_buffer = io.StringIO()
    df.to_csv(csv_buffer, index=False)

    # Extract the bucket and the relative path from the provided S3 URI
    parsed_url = urlparse(key)
    bucket = parsed_url.netloc
    relative_path = parsed_url.path.lstrip("/")

    # Upload CSV to S3
    s3_client.put_object(Bucket=bucket, Key=relative_path, Body=csv_buffer.getvalue())


def read_s3_object_as_dataframe(key, s3_client):
    parsed_url = urlparse(key)
    bucket = parsed_url.netloc
    relative_path = parsed_url.path.lstrip("/")
    response = s3_client.get_object(Bucket=bucket, Key=relative_path)
    content = response["Body"].read().decode("utf-8")
    df = pd.read_csv(StringIO(content))
    return df


class AwsS3:
    def __init__(self, params, cross_account_arn=None):
        self.params = params
        self.cross_account_arn = cross_account_arn
        logger.debug(
            f"Create AwsS3 instance with {self.params} and x-account {self.cross_account_arn}"
        )

    def upload_file(self, file_name, bucket, object_name=None):
        """Upload a file to an S3 bucket

        :param file_name: File to upload
        :param bucket: Bucket to upload to
        :param object_name: S3 object name. If not specified then file_name is used
        :return: True if file was uploaded, else False
        """

        # If S3 object_name was not specified, use file_name
        if object_name is None:
            object_name = file_name

        # Upload the file
        # s3_client = boto3.client("s3")
        s3_client = get_s3_client(self.cross_account_arn)
        try:
            s3_client.upload_file(file_name, bucket, object_name)
        except botocore.exceptions.ClientError as e:
            logger.error(e)
            return False
        return True

    def s3_file_exists(self, s3_file_name):
        s3 = get_s3_resource(self.cross_account_arn)
        bucket = s3.Bucket(self.params["outputBucket"])
        key = (
            self.params["outputFolder"] + "/" + s3_file_name
            if self.params["outputFolder"] != ""
            else s3_file_name
        )
        objs = list(bucket.objects.filter(Prefix=key))
        if len(objs) > 0 and objs[0].key == key:
            return True
        else:
            return False

    def get_dataframe_from_s3(self, s3_bucket, s3_key, col_names=None, escapechar=None):
        try:
            s3_client = get_s3_client(self.cross_account_arn)
            obj = s3_client.get_object(Bucket=s3_bucket, Key=s3_key)
            if col_names is not None:
                dataframe = pd.read_csv(
                    obj["Body"],
                    delimiter=",",
                    names=col_names,
                    header=None,
                    escapechar=escapechar,
                )
            else:
                dataframe = pd.read_csv(
                    obj["Body"], delimiter=",", escapechar=escapechar
                )
            return dataframe
        except Exception as e:
            logger.error(f"Error in get_dataframe_from_s3. Details: {e}")
        return None

    def dataframe_to_s3(self, dataframe, s3_bucket, s3_key):
        try:
            s3_client = get_s3_client(self.cross_account_arn)
            csv_buf = StringIO()
            dataframe.to_csv(csv_buf, header=True, index=False)
            csv_buf.seek(0)
            s3_client.put_object(Bucket=s3_bucket, Body=csv_buf.getvalue(), Key=s3_key)
            return True
        except Exception as e:
            logger.error(f"Error in save_dataframe. Details: {e}")
        return False

    def get_dataframe(self, query, s3_file_name, force_query=False):
        try:
            s3_file_exists = True
            if force_query or not self.s3_file_exists(s3_file_name):
                s3_file_exists = self.run_query_on_s3(query, s3_file_name)
            if s3_file_exists:
                bucket = self.params["outputBucket"]
                key = (
                    self.params["outputFolder"] + "/" + s3_file_name
                    if self.params["outputFolder"] != ""
                    else s3_file_name
                )
                return self.get_dataframe_from_s3(bucket, key)
            else:
                return None
        except Exception as e:
            logger.error(f"Error in get_dataframe. Details: {e}")
        return None

    def get_res_on_s3(self, query, s3_file_name, force_query=False):
        try:
            s3_file_exists = True
            if force_query or not self.s3_file_exists(s3_file_name):
                s3_file_exists = self.run_query_on_s3(query, s3_file_name)
            if s3_file_exists:
                return (
                    self.params["outputLocation"] + s3_file_name,
                    self.params["outputBucket"],
                    self.params["outputFolder"] + "/" + s3_file_name,
                )
            else:
                return None
        except Exception as e:
            traceback.print_exc()
            logger.error(f"Error in get_res_on_s3. Details: {e}")
        return None

    def run_query_on_s3(self, query, new_s3_file_name):
        """Trigger athena query then copy query result to specified folder of output bucket."""
        cursor = self.get_athena_cursor()
        logger.debug(f"prepare to run {query}")
        cursor.execute(query)
        result_output_location = cursor.output_location
        logger.info(f"query out put location: {result_output_location}")
        query_result_filename = result_output_location.split("/")[-1]
        if query_result_filename:
            logger.debug(f"File name of query result: {query_result_filename}")
            s3 = get_s3_resource(self.cross_account_arn)
            copy_source = {
                "Bucket": self.params["bucket"],
                "Key": query_result_filename,
            }
            bucket = s3.Bucket(self.params["outputBucket"])
            if self.params["outputFolder"] != "":
                bucket.copy(
                    copy_source, f"{self.params['outputFolder']}/{new_s3_file_name}"
                )
            else:
                bucket.copy(copy_source, new_s3_file_name)
            return True
        else:
            logger.error("failed to run query on Athena!")
            return False

    def athena_to_s3(self, session, query, max_execution=5000):
        """Execute athena query. Return query result filename when succeed"""
        client = session.client("athena", region_name=self.params["region"])
        execution = self.athena_query(client, query)
        execution_id = execution["QueryExecutionId"]
        state = "RUNNING"
        while max_execution > 0 and state in ["RUNNING", "QUEUED"]:
            max_execution = max_execution - 1
            logger.debug(max_execution)
            response = client.get_query_execution(QueryExecutionId=execution_id)

            if (
                "QueryExecution" in response
                and "Status" in response["QueryExecution"]
                and "State" in response["QueryExecution"]["Status"]
            ):
                state = response["QueryExecution"]["Status"]["State"]
                logger.debug(state)
                if state == "FAILED":
                    logger.debug(response)
                    return False
                elif state == "SUCCEEDED":
                    s3_path = response["QueryExecution"]["ResultConfiguration"][
                        "OutputLocation"
                    ]
                    filename = re.findall(".*\\/(.*)", s3_path)[0]
                    return filename
            time.sleep(5)
        return False

    def athena_query(self, client, query):

        response = client.start_query_execution(
            QueryString=query,
            QueryExecutionContext={"Database": self.params["database"]},
            # TODO: Set query result location once MOSAIC20-1321 is completed.
            # ResultConfiguration={"OutputLocation": "s3://" + self.params["bucket"]}
            WorkGroup=os.environ["AWS_ATHENA_WORK_GROUP"],
        )
        return response

    def get_athena_cursor(self):
        if not self.cross_account_arn:
            logger.info(
                f'Create non cross account athena cursor. query output: {self.params["bucket"]}'
            )
            # Uses mosaic2 IAM role to execute query.
            return connect(
                s3_staging_dir=f"s3://{self.params['bucket']}/",
                region_name=self.params["region"],
            ).cursor()
        else:
            logger.info(
                f"Create cross account athena cursor. ARN: {self.cross_account_arn}."
                f' query output: {self.params["bucket"]}'
            )
            # Assume IAM role to execute query in cdc-analytics.
            return connect(
                work_group="primary",
                s3_staging_dir=f"s3://{self.params['bucket']}/",
                role_arn=self.cross_account_arn,
                role_session_name="PyAthena-session",
                duration_seconds=3600,
                region_name=self.params["region"],
            ).cursor()
