import logging
import os

from pyathena import connect

from mycroft.common import validate_date
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def process_daily_intermediate_table(**kwargs):
    """
    Create database and 'proc_hash_cert_daily_alert_dev_count' table
    if dag_run json is specified while triggering DAG execution.
    { "initialize_intermediate_table": "true", "initial_date": "YYYY-MM-DD"}

    Update table 'proc_hash_cert_daily_alert_dev_count' otherwise
    """

    try:
        region = kwargs["region"]
        query_output_s3_bucket = kwargs["query_output_s3_bucket"]
        flink_db_name = kwargs["from_database"]
        daily_intermediate_db_name = kwargs["to_database"]
        database_s3_bucket = kwargs["database_s3_bucket"]
        legacy_env = kwargs["legacy_env"]
        msc1_role_arn = kwargs["msc1_role_arn"]
        logger.info(f"Daily update intermediate table on {region}, DB: {flink_db_name}")
        if (
            "dag_run" in kwargs
            and kwargs["dag_run"].conf is not None
            and kwargs["dag_run"].conf.get("initialize_intermediate_table")
        ):
            initial_date = kwargs["dag_run"].conf.get("initial_date")
            initialization(
                region,
                query_output_s3_bucket,
                flink_db_name,
                database_s3_bucket,
                daily_intermediate_db_name,
                initial_date,
            )
        else:
            execution_date = kwargs["execution_date"]
            daily_update_task(
                region,
                query_output_s3_bucket,
                flink_db_name,
                daily_intermediate_db_name,
                execution_date,
                legacy_env,
                msc1_role_arn,
            )
    except KeyError as e:
        logger.error("Error in finding the key in kwargs. Error details: %s", str(e))
        raise Exception("Failed to process_daily_intermediate_table.")


def initialization(
    region,
    query_output_s3_bucket,
    flink_db_name,
    database_s3_bucket,
    daily_intermediate_db_name,
    initial_date,
):
    """
    Initialize proc_hash_cert_alert database and  proc_hash_cert_daily_alert_dev_count table,
    Create proc_hash_cert_alert database,
    Initialize proc_hash_cert_daily_alert_dev_count table,
    Use a CTAS query to create the table.
    Only required in new Pop env.
    """
    try:
        validate_date(initial_date)
        logger.debug("Initialize proc_hash_cert_daily_alert_dev_count table")
        cursor = connect(
            s3_staging_dir=f"s3://{query_output_s3_bucket}/",
            region_name=region,
        ).cursor()

        create_database_command = f"""
        CREATE DATABASE IF NOT EXISTS `{daily_intermediate_db_name}`
        LOCATION '{database_s3_bucket}'
        """
        logger.debug(create_database_command)
        cursor.execute(create_database_command)

        create_table_command = f"""
        CREATE table {daily_intermediate_db_name}.proc_hash_cert_daily_alert_dev_count
        WITH (format='ORC', partitioned_by=array['event_date', 'env']) AS
        WITH t1 AS
            (SELECT  augmentedbehaviorevent_behavior_processname AS proc_path,
                     augmentedbehaviorevent_behavior_processhash AS proc_sha256hash,
                     eventcontextinput_vtinfo_signatureinfo_signedby AS cert_signedby,
                     eventcontextinput_vtinfo_signatureinfo_signingauthority AS cert_authority,
                     deviceid AS dev_id,
                     augmentedbehaviorevent_behavior_header_devicetype AS dev_type,
                     environmentname AS env,
                     CAST(organizationid AS Integer) AS org_id,
                     CAST(augmentedbehaviorevent_groupid AS Integer) AS policy_id,
                     event_date
            FROM {flink_db_name}.flink_cbd
            WHERE event_date = '{initial_date}' AND try(CAST(organizationid AS INTEGER)) IS NOT NULL
            AND try(CAST(augmentedbehaviorevent_groupid AS INTEGER)) IS NOT NULL
            AND deviceeventlog_policystate IN ('APPLIED', 'NOT_APPLIED')
            AND deviceeventlog_policystate = 'APPLIED'
                    AND ((augmentedbehaviorevent_behavior_header_code = 'REQ_CREATE_PROCESS_BEHAVIOR'
                    AND deviceeventlog_targeteffectivereputation IN ('ADAPTIVE_WHITE_LIST','NOT_LISTED','RESOLVING') )
                    OR (augmentedbehaviorevent_behavior_header_code != 'REQ_CREATE_PROCESS_BEHAVIOR'
                    AND deviceeventlog_processeffectivereputation IN ('ADAPTIVE_WHITE_LIST','NOT_LISTED','RESOLVING'))))
        SELECT   t1.proc_path AS proc_path,
                 t1.proc_sha256hash AS proc_sha256hash,
                 t1.cert_signedby AS cert_signedby,
                 t1.cert_authority AS cert_authority,
                 t1.org_id AS org_id,
                 t1.policy_id as policy_id,
                 t1.dev_type AS dev_type,
                 CAST(APPROX_SET(t1.dev_id) AS VARBINARY) AS alert_device_count_hll_sketch,
                 count(t1.dev_id) AS alert_cnt,
                 t1.event_date AS event_date,
                 t1.env AS env
        FROM t1
        GROUP BY  1, 2, 3, 4, 5, 6, 7, 10, 11
        """
        logger.debug(create_table_command)
        cursor.execute(create_table_command)
    except Exception as e:
        logger.error(
            "Error in initializing proc_hash_cert_daily_alert_dev_count table. Error details: %s",
            str(e),
        )


def daily_update_task(
    region,
    query_output_s3_bucket,
    flink_db_name,
    daily_intermediate_db_name,
    execution_date,
    legacy_env,
    msc1_role_arn,
):
    """
    Insert a single day data into proc_hash_cert_daily_alert_dev_count table
    Suppose execution date is 2020-02-02, then it will insert data of date 2020-02-01
    """
    date = execution_date.add(days=-1).to_date_string()
    try:
        validate_date(date)
        logger.info(
            f"Inserting data on [{date}], from [{flink_db_name}] in region: [{region}] env: [{legacy_env}] "
            f"to db {daily_intermediate_db_name}"
        )
        if legacy_env == "":
            # Uses mosaic2 IAM role to execute query.
            cursor = connect(
                s3_staging_dir=f"s3://{query_output_s3_bucket}/",
                region_name=region,
            ).cursor()
            command = get_pop_command(flink_db_name, date)
        else:
            # Assume IAM role to execute query in cdc-analytics.
            cursor = connect(
                work_group="primary",
                role_arn=msc1_role_arn,
                role_session_name="PyAthena-session",
                duration_seconds=3600,
                s3_staging_dir=f"s3://{query_output_s3_bucket}/",
                region_name=region,
            ).cursor()
            command = get_command(legacy_env, flink_db_name, date)

        logger.debug(f"Prepare to execute {command}")
        cursor.execute(command)
    except ValueError as e:
        logger.error("Error in the value. Error details: %s", str(e))
        raise Exception("Failed to process_daily_intermediate_table.")
    except Exception as e:
        logger.error(
            "Error in inserting data. Date: %s. Error details: %s",
            date,
            str(e),
        )
        raise Exception("Failed to process_daily_intermediate_table.")


def get_command(legacy_env, database, date):
    if legacy_env == "dev01":
        legacy_env = "dev"
    elif legacy_env == "test03":
        legacy_env = "stage"
    elif legacy_env == "prod05":
        legacy_env = "prod"
    return f"""INSERT INTO proc_hash_cert_alert_db.proc_hash_cert_daily_alert_dev_count(
                    proc_path,
                    proc_sha256hash,
                    cert_signedby,
                    cert_authority,
                    org_key,
                    policy_id,
                    dev_type,
                    alert_device_count_hll_sketch,
                    alert_cnt,
                    event_date,
                    org_id,
                    env
                )
                SELECT
                    subquery.proc_path AS proc_path,
                    subquery.proc_sha256hash AS proc_sha256hash,
                    subquery.cert_signedby AS cert_signedby,
                    subquery.cert_authority AS cert_authority,
                    subquery.org_key AS org_key,
                    subquery.policy_id AS policy_id,
                    subquery.dev_type AS dev_type,
                    CAST(APPROX_SET(subquery.dev_id) AS VARBINARY) AS alert_device_count_hll_sketch,
                    count(subquery.dev_id) AS alert_cnt,
                    subquery.event_date AS event_date,
                    NULL AS org_id,
                    NULL AS env
                FROM (
                    SELECT
                        actor_process_path AS proc_path,
                        actor_process_hash_sha256 AS proc_sha256hash,
                        CASE
                            WHEN event_type = 'TYPE_FILE'
                                AND cardinality(filemod_op_target_digital_signatures_publisher) > 0
                                AND filemod_op_target_digital_signatures_publisher [1] != ''
                                    THEN filemod_op_target_digital_signatures_publisher [1]
                            WHEN event_type = 'TYPE_MOD_LOAD'
                                AND modload_op_target_digital_signature1_publisher != ''
                                    THEN modload_op_target_digital_signature1_publisher
                            WHEN event_type = 'TYPE_PROCESS'
                                AND cardinality(actor_process_digital_signatures_publisher) > 0
                                AND actor_process_digital_signatures_publisher [1] != ''
                                    THEN actor_process_digital_signatures_publisher [1]
                            ELSE ''
                        END AS cert_signedby,
                        CASE
                            WHEN event_type = 'TYPE_FILE'
                                AND cardinality(filemod_op_target_digital_signatures_issuer) > 0
                                AND filemod_op_target_digital_signatures_issuer [1] != ''
                                    THEN filemod_op_target_digital_signatures_issuer [1]
                            WHEN event_type = 'TYPE_MOD_LOAD'
                                AND modload_op_target_digital_signature1_issuer != ''
                                    THEN modload_op_target_digital_signature1_issuer
                            WHEN event_type = 'TYPE_PROCESS'
                                AND cardinality(actor_process_digital_signatures_issuer) > 0
                                AND actor_process_digital_signatures_issuer [1] != ''
                                    THEN actor_process_digital_signatures_issuer [1]
                        ELSE ''
                        END AS cert_authority,
                        device_id AS dev_id,
                        device_os_type AS dev_type,
                        org_key AS org_key,
                        CAST(device_policy_id AS Integer) AS policy_id,
                        event_date,
                        row_number() over (PARTITION by event_id ORDER BY device_timestamp DESC) AS seqnum
                    FROM {database}.flink_es_event_{legacy_env}
                    WHERE event_date = '{date}'
                        AND org_key IS NOT NULL
                        AND try(CAST(device_policy_id AS INTEGER)) IS NOT NULL
                        AND sensor_action != 'ACTION_ALLOW'
                        AND (
                                (
                                  op_action != 'ACTION_CREATE_PROCESS'
                                  AND actor_process_effective_reputation IN (
                                    'REP_ADAPTIVE',
                                    'REP_NOT_LISTED',
                                    'REP_RESOLVING'
                                  )
                                )
                                OR (
                                  (
                                    op_action = 'ACTION_CREATE_PROCESS'
                                    AND process_op_target_effective_reputation IN (
                                      'REP_ADAPTIVE',
                                      'REP_NOT_LISTED',
                                      'REP_RESOLVING'
                                    )
                                  )
                                )
                              )
                ) AS subquery
                WHERE seqnum = 1
                GROUP BY  1, 2, 3, 4, 5, 6, 7, 10, 11, 12"""


def get_pop_command(flink_db_name, date):
    """
    UKPoP specific command to update daily count table using mosaic2 P9 table.
    The database containing tables named
    with the prefix 'flink_' is distinct from the database that includes tables prefixed with 'policy_' in vdp cluster.
    """
    flink_cbd_db_name = os.getenv("FLINK_CBD_DB_NAME", "")
    return f"""
        INSERT INTO proc_hash_cert_alert_db.proc_hash_cert_daily_alert_dev_count
        With t1 AS
            (SELECT augmentedbehaviorevent_behavior_processname AS proc_path,
                    augmentedbehaviorevent_behavior_processhash AS proc_sha256hash,
                    eventcontextinput_vtinfo_signatureinfo_signedby AS cert_signedby,
                    eventcontextinput_vtinfo_signatureinfo_signingauthority AS cert_authority,
                    deviceid AS dev_id,
                    augmentedbehaviorevent_behavior_header_devicetype AS dev_type,
                    environmentname AS env,
                    CAST(organizationid AS Integer) AS org_id,
                    CAST(augmentedbehaviorevent_groupid AS Integer) AS policy_id,
                    event_date
            FROM {flink_cbd_db_name}.flink_cbd
            WHERE event_date = '{date}' AND try(CAST(organizationid AS INTEGER)) IS NOT NULL
            AND try(CAST(augmentedbehaviorevent_groupid AS INTEGER)) IS NOT NULL
            AND deviceeventlog_policystate IN ('APPLIED', 'NOT_APPLIED')
            AND deviceeventlog_policystate = 'APPLIED'
                    AND ((augmentedbehaviorevent_behavior_header_code = 'REQ_CREATE_PROCESS_BEHAVIOR'
                    AND deviceeventlog_targeteffectivereputation IN ('ADAPTIVE_WHITE_LIST','NOT_LISTED','RESOLVING') )
                    OR (augmentedbehaviorevent_behavior_header_code != 'REQ_CREATE_PROCESS_BEHAVIOR'
                    AND deviceeventlog_processeffectivereputation IN ('ADAPTIVE_WHITE_LIST','NOT_LISTED','RESOLVING'))))
        SELECT  t1.proc_path AS proc_path,
                t1.proc_sha256hash AS proc_sha256hash,
                t1.cert_signedby AS cert_signedby,
                t1.cert_authority AS cert_authority,
                t1.org_id AS org_id,
                t1.policy_id AS policy_id,
                t1.dev_type AS dev_type,
                CAST(APPROX_SET(t1.dev_id) AS VARBINARY) AS alert_device_count_hll_sketch,
                count(t1.dev_id) AS alert_cnt,
                t1.event_date AS event_date,
                t1.env AS env,
                '' AS org_key
        FROM t1
        GROUP BY  1, 2, 3, 4, 5, 6, 7, 10, 11"""
