#!/usr/bin/env python
# coding: utf-8
import os

import numpy as np
import pandas as pd

from mycroft.common import gen_signature
from mycroft.common.aws_s3 import (
    AwsS3,
    get_s3_client,
    read_s3_object_as_dataframe,
    upload_dataframe_to_s3,
)
from mycroft.common.logging import get_logger

LOGLEVEL = os.environ.get("LOGLEVEL", "WARNING").upper()
logger = get_logger(LOGLEVEL)


def get_existing_it_tools_rules(
    s3_output_bucket,
    s3_output_folder,
    s3_output_filename,
    region="us-east-1",
    database="athena",
    table="policy_org_reputation_override",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
    cross_account_arn=None,
):
    params = {
        "region": region,
        "database": database,
        "bucket": query_bucket,
        "outputBucket": s3_output_bucket,
        "outputFolder": s3_output_folder,
        "outputLocation": f"s3://{s3_output_bucket}/{s3_output_folder}/",
    }
    aws_data = AwsS3(params, cross_account_arn=cross_account_arn)
    query = f"""SELECT lower(value1) rule_path,
                                orgid AS org_id, env
                FROM {database}.{table}
                WHERE TYPE = 'IT_TOOL'
                    AND authlisttype = 'WHITE_LIST'
                    AND value1 LIKE '%\\%'
                    AND value1 NOT LIKE '%,%'
                    AND value1 NOT IN ('*\\*',
                                        '**\\**',
                                        '**',
                                        '*') """
    s3_file_path, s3_file_bucket, s3_file_key = aws_data.get_res_on_s3(
        gen_signature(query, "policy-automation"),
        s3_output_filename,
        force_query=True,
    )
    return s3_file_key


def gen_it_tools_reputation_pool(
    prevalence_thresh,
    s3_output_bucket,
    s3_output_folder,
    s3_output_filename,
    region="us-east-1",
    database="athena",
    table="policy_org_reputation_override",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
):
    params = {
        "region": region,
        "database": database,
        "bucket": query_bucket,
        "outputBucket": s3_output_bucket,
        "outputFolder": s3_output_folder,
        "outputLocation": f"s3://{s3_output_bucket}/{s3_output_folder}/",
    }
    aws_data = AwsS3(params)
    query = f""" WITH t1 AS
                (SELECT DISTINCT lower(value1) application_value,
                                count(DISTINCT orgid) org_cnt
                FROM {database}.{table}
                WHERE TYPE = 'IT_TOOL'
                    AND authlisttype = 'WHITE_LIST'
                    AND value1 LIKE '%\\%'
                    AND value1 NOT LIKE '%,%'
                    AND value1 NOT IN ('*\\*',
                                        '**\\**',
                                        '**',
                                        '*')
                GROUP BY 1
                ORDER BY 2 DESC)
            SELECT t1.*
            FROM t1
            WHERE t1.org_cnt > {prevalence_thresh} """
    s3_file_path, s3_file_bucket, s3_file_key = aws_data.get_res_on_s3(
        gen_signature(query, "policy-automation"),
        s3_output_filename,
        force_query=True,
    )
    return s3_file_path


def get_org_cnt_value2(aws_data, it_tools_regex, database, table, force_query):
    query = f"""SELECT orgid, lower(value2) AS value2 FROM {database}.{table} WHERE regexp_like(lower(value1), '{it_tools_regex}') AND TYPE = 'IT_TOOL'"""
    dataframe = aws_data.get_dataframe(
        gen_signature(query, "policy-automation"), "it_tool_rule_temp.csv", force_query
    )
    if not dataframe.empty:
        org_cnt = dataframe["orgid"].nunique()
        dataframe["value2"] = dataframe["value2"].astype(str).str.lower()
        df = (
            dataframe.groupby("value2")
            .agg({"orgid": "size"})
            .sort_values("orgid", ascending=False)
            .reset_index()
            .rename(columns={"orgid": "count"})
        )
        true_value = 0
        false_value = 0
        if "true" in df["value2"]:
            true_value = df[df["value2"] == "true"]["count"]
        if "false" in df["value2"]:
            false_value = df[df["value2"] == "false"]["count"]
        if true_value >= false_value:
            value2 = "True"
        else:
            value2 = "False"
        return org_cnt, value2
    else:
        return 0, "True"


def gen_params_for_it_tools_reputation_pool(
    static_it_tools_rules_path,
    s3_query_bucket,
    s3_output_filename,
    region="us-east-1",
    database="athena",
    table="policy_org_reputation_override",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
    cross_role_arn=None,
):
    params = {
        "region": region,
        "database": database,
        "bucket": query_bucket,
        "outputBucket": s3_query_bucket,
        "outputFolder": "",
        "outputLocation": f"s3://{s3_query_bucket}/",
    }
    force_query = True
    aws_data = AwsS3(params, cross_role_arn)
    data = []
    client = get_s3_client(cross_account_arn=cross_role_arn)
    df = read_s3_object_as_dataframe(static_it_tools_rules_path, client)
    # df = pd.read_csv(static_it_tools_rules_path, delimiter=",", header=None)
    for index, row in df.iterrows():
        try:
            org_cnt, value2 = get_org_cnt_value2(
                aws_data, row[1], database, table, force_query
            )
            data.append([row[0], org_cnt, value2])
        except Exception as e:
            logger.error(f"Cannot get org_cnt, value2 for {row[1]}; Error: {e}")
    pool_dataframe = pd.DataFrame(
        np.array(data), columns=["application_value", "org_cnt", "value2"]
    )
    upload_dataframe_to_s3(pool_dataframe, s3_output_filename, client)
    # pool_dataframe.to_csv(s3_output_filename, index=False, header=False)
    return s3_output_filename


def gen_permission_rules_pool(
    prevalence_thresh,
    s3_output_bucket,
    s3_output_folder,
    s3_output_filename,
    region="us-east-1",
    database="athena",
    table="policy",
    query_bucket="aws-athena-query-results-prod-us-east-1-shannon",
):
    """
    Generates permission rules with high prevalence from all CB orgs.

    :param session: Boto3 session
    :param prevalence_thresh: Prevalence threshold can be number of orgs that have the rule
    :return: The s3 path of the generated csv file and status of the query
    """
    params = {
        "region": region,
        "database": database,
        "bucket": query_bucket,
        "outputBucket": s3_output_bucket,
        "outputFolder": s3_output_folder,
        "outputLocation": f"s3://{s3_output_bucket}/{s3_output_folder}/",
    }
    aws_data = AwsS3(params)
    query = f""" WITH t1 AS
                        (SELECT lower(application_value) application_value,
                                operation,
                                action,
                                count(DISTINCT org_id) org_cnt
                        FROM {database}.{table}
                        WHERE application_value LIKE '%\\%'
                            AND application_value NOT LIKE '%,%'
                            AND application_value NOT IN ('*\\*',
                                                        '**\\**',
                                                        '**',
                                                        '*')
                            AND (action = 'ALLOW'
                                OR action = 'IGNORE')
                        GROUP BY 1,
                                 2,
                                 3
                        ORDER BY 4 DESC)
                      SELECT t1.*
                      FROM t1
                      WHERE t1.org_cnt > {prevalence_thresh} """
    s3_file_path, s3_file_bucket, s3_file_key = aws_data.get_res_on_s3(
        gen_signature(query, "policy-automation"),
        s3_output_filename,
        force_query=True,
    )
    return s3_file_path
